import re

from lxml import etree

from . constants import EMAIL_REGEXP
from . indicepa import get_indicepa_by_ipacode


class SpidSpMetadataCheckPublic(object):

    def test_Contacts_PubPriv(self, contact_type="other"):
        entity_desc = self.doc.xpath('//ContactPerson')
        self._assertTrue(entity_desc, 'ContactPerson MUST be present')

        if entity_desc:
            self._assertTrue(
                entity_desc[0].attrib.get('contactType'),
                (f'Missing contactType in {entity_desc[0].attrib}: '
                 'The contactType attribute MUST be present - TR pag. 19')
            )
            self._assertTrue(
                entity_desc[0].get('contactType'),
                'The contactType attribute MUST have a value - TR pag. 19'
            )
            self._assertTrue(
                entity_desc[0].get('contactType') == 'other',
                'The contactType must be "other" - TR pag. 19',
                description = entity_desc[0].get('contactType')
            )

        others = self.doc.xpath(
            f'//ContactPerson[@contactType="{contact_type}"]')
        self._assertTrue(
            len(others) == 1,
            f'Only one ContactPerson element of contactType "{contact_type}" MUST be present',
            description = others
        )

        return self.is_ok(f'{self.__class__.__name__}.test_Contacts_PubPriv')

    def test_Extensions_PubPriv(self):
        _conts = self.doc.xpath('//ContactPerson')

        for cont in _conts:
            ext_cnt = 0
            for child in cont.getchildren():
                if child.tag == 'Extension':
                    ext_cnt += 1

            self._assertFalse(
                ext_cnt > 1,
                'Only one Extensions element inside ContactPerson element MUST be present',
                description = etree.tostring(cont).decode()
            )

        orgs = self.doc.xpath('//EntityDescriptor/Organization/OrganizationName')
        if len(orgs) >= 1:
            org = orgs[0]
            company = self.doc.xpath('//ContactPerson/CompanyName')
            if company:
                company = company[0]

                self._assertTrue(
                    company.text,
                    'If the Company element is present it MUST have a value',
                    description = company,
                )

                self._assertTrue(
                    company.text == org.text,
                    'If the Company element if present it MUST be equal to OrganizationName',
                    description = (company.text, org.text)
                )

        return self.is_ok(f'{self.__class__.__name__}.test_Extensions_PubPriv')

    def test_contactperson_email(self,
                                 email_xpath='//ContactPerson/EmailAddress'):
        email = entity_desc = self.doc.xpath(email_xpath)
        self._assertTrue(
            email,
            f'The {email_xpath} element MUST be present',
            description = email,
        )
        if email:
            self._assertTrue(
                email[0].text,
                f'The {email_xpath} element MUST have a value',
                description = email[0],
            )
            self._assertTrue(
                re.match(EMAIL_REGEXP, email[0].text),
                f'The {email_xpath} element MUST be a valid email address',
                description = email[0],
            )
        return self.is_ok(f'{self.__class__.__name__}.test_contactperson_email-{email_xpath}')

    def test_contactperson_phone(self, phone_xpath='//ContactPerson/TelephoneNumber'):
        phone = entity_desc = self.doc.xpath(phone_xpath)
        if phone:
            phone = phone[0].text
            self._assertTrue(
                phone,
                f'The {phone_xpath} element MUST have a value',
            )
            self._assertTrue(
                (' ' not in phone),
                f'The {phone_xpath} element MUST not contain spaces',
                description = phone,
            )
            self._assertTrue(
                (phone[0:3] == '+39'),
                f'The {phone_xpath} element MUST start with "+39"',
                description = phone,
                level = 'warning'
            )

        return self.is_ok(f'{self.__class__.__name__}.test_contactperson_phone')


    def test_Contacts_IPACode(self):
        entity_desc = self.doc.xpath('//ContactPerson')

        if self.production:
            ipacode = self.doc.xpath('//ContactPerson/Extensions/IPACode')
            self._assertTrue(
                ipacode,
                'The IPACode element MUST be present',
            )
            if ipacode:
                ipacode = ipacode[0]
                self._assertTrue(
                    ipacode.text,
                    'The IPACode element MUST have a value',
                )
                self._assertTrue(
                    ipacode.text,
                    'The IPACode element MUST have a value',
                )
                self._assertTrue(
                    get_indicepa_by_ipacode(ipacode.text)[0] == 1,
                    'The IPACode element MUST have a valid value present on IPA',
                )

        return self.is_ok(f'{self.__class__.__name__}.test_Contacts_IPACode')

    def test_extensions_public_private(self, ext_type="Public"):
        ext_type_not = "Private" if ext_type == "Public" else "Public"

        # only if other, billing doesn't have any Private element in it!
        ctype = self.doc.xpath(f'//ContactPerson[@contactType="other"]/Extensions/{ext_type.title()}')
        self._assertTrue(
            ctype,
            f'Missing ContactPerson/Extensions/{ext_type.title()}, this element MUST be present',
        )
        if ctype:
            self._assertFalse(
                ctype[0].text,
                f'The {ext_type.title()} element MUST be empty',
            )

        ctype = self.doc.xpath(f'//ContactPerson/Extensions/{ext_type_not.title()}')
        self._assertFalse(
            ctype,
            f'The {ext_type_not.title()} element MUST not be present',
        )
        return self.is_ok(f'{self.__class__.__name__}.test_extentions_public')

    def test_Contacts_VATFC(self, private=False):
        entity_desc = self.doc.xpath('//ContactPerson')

        vats = self.doc.xpath('//ContactPerson/Extensions/VATNumber')
        self._assertTrue(
            (len(vats) <= 1),
            'only one VATNumber element must be present',
            description = vats
        )
        if vats:
            self._assertTrue(
                vats[0].text,
                'The VATNumber element MUST have a value',
            )

        fcs = self.doc.xpath('//ContactPerson/Extensions/FiscalCode')
        if fcs:
            self._assertTrue(
                (len(fcs) == 1),
                'only one FiscalCode element must be present',
                description = fcs
            )
            fc = fcs[0]
            self._assertTrue(
                fc.text,
                'The FiscalCode element MUST have a value',
            )
        if private and not len(fcs) and not len(vats):
            self._assertTrue(
                False,
                'If the VATNumber is not present, the FiscalCode element MUST be present',
            )

        return self.is_ok(f'{self.__class__.__name__}.test_Contacts_VATFC')



# Copyright 2018 AgID - Agenzia per l'Italia Digitale
#
# Licensed under the EUPL, Version 1.2 or - as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence").
#
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at:
#
#    https://joinup.ec.europa.eu/software/page/eupl
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# Licence for the specific language governing permissions and limitations
# under the Licence.

import base64
import re
import OpenSSL


def _dump_pem(xml_elem, data_dir, context, use):
    pem = []
    n = 78

    b64 = re.sub(r'[\s]', '', xml_elem.text)
    pem.append('-----BEGIN CERTIFICATE-----')
    [pem.append(b64[i:i + n]) for i in range(0, len(b64), n)]
    pem.append('-----END CERTIFICATE-----')

    x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1,
                                           base64.b64decode(b64))

    dgst = x509.digest('sha256').decode('utf-8').replace(':', '')

    fname = ('%s/%s.%s.%s.pem') % (data_dir, dgst[0:16], context, use)

    with open(fname, 'w') as f:
        f.write('\n'.join(pem))
        f.close()

    return fname


def dump_metadata_pem(xml_elem, context, use, data_dir):
    return _dump_pem(xml_elem, data_dir, '%s.metadata' % context, use)


def dump_request_pem(xml_elem, context, use, data_dir):
    return _dump_pem(xml_elem, data_dir, '%s.request' % context, use)


def dump_response_pem(xml_elem, context, use, data_dir):
    return _dump_pem(xml_elem, data_dir, '%s.response' % context, use)


def dump_assertion_pem(xml_elem, context, use, data_dir):
    return _dump_pem(xml_elem, data_dir, '%s.assertion' % context, use)



import logging
import re

from pathlib import Path


BASE_DIR = Path(__file__).resolve().parent
logger = logging.getLogger(__name__)


class AbstractSpidCheck(object):
    def __init__(self, *args, **kwargs):
        self.results = []
        self.errors = []
        self.logger = logger
        self.error_counter = 0
        self.verify_ssl = kwargs.get('verify_ssl', False)
        self.category = ''

    def report_to_dict(self):
        res = {
            self.category: {
                self.__class__.__name__: self.results
            }
        }

        return res

    def is_ok(self, msg):
        if not self.error_counter:
            self.handle_result('info', f"{msg}")
            return True
        else:
            self.error_counter = 0
            return False

    def handle_result(self,
                      level: str,
                      title: str, description: str = '',
                      traceback: str = None):
        msg = f'{title}'
        getattr(self.logger, level, 'debug')(msg)
        value = f'{description}' if not traceback else f'{description}: {traceback }'
        if level not in ('error', 'debug', 'critical', 'warning'):
            # here report as json
            self.results.append(
                {
                    "result": "success",
                    "test": title,
                    # "value": value
                }
            )
        elif level in ('error', 'critical'):
            self.handle_error(title,
                              description,
                              traceback)
        elif level == 'warning':
            self.results.append(
                {
                    "result": "warning",
                    "test": title,
                    "value": value,
                }
            )

    def handle_error(self, error_message, description='',
                     traceback: str = None):
        getattr(self.logger, 'error')(error_message)
        self.error_counter += 1
        # here report as json
        value = f'{description}' if not traceback else f'{description}: {traceback }'
        data = {
            "result": "failure",
            "test": error_message,
            "value": value
        }
        self.errors.append(data)
        self.results.append(data)

    def _assert(self, check:bool, error_message:str,
                description='', traceback: str = None,
                level: str = 'info'):
        if not check and level == 'info':
            self.handle_error(error_message, description, traceback)
        elif not check and level == 'warning':
            self.handle_result(level, f"{error_message}", description, traceback)
        else:
            level = 'info' if level in ('warning',) else level
            self.handle_result(level, f"{error_message}", description, traceback)

    def _assertTrue(self, check, error_message,
                    description='', traceback: str = None,
                    level: str = 'info'):
        self._assert(check, error_message, description, traceback, level)

    def _assertFalse(self, check, error_message,
                     description='', traceback: str = None,
                     level: str = 'info'):
        self._assert(not check, error_message, description, traceback, level)

    def _assertIsNotNone(self, check, error_message,
                         description='', traceback: str = None,
                         level: str = 'info'):
        self._assert(check, error_message, description, traceback, level)

    def _assertIn(self, first, second, error_message,
                  description='', traceback: str = None,
                  level: str = 'info'):
        self._assert((first in second),
                     error_message, description, traceback, level)

    def _assertGreaterEqual(self, first, second, error_message,
                            description='', traceback: str = None,
                            level: str = 'info'):
        self._assert((first >=second),
                     error_message, description, traceback, level)

    def _assertGreater(self, first, second, error_message,
                       description='', traceback: str = None,
                       level: str = 'info'):
        self._assert((first > second),
                     error_message, description, traceback, level)

    def _assertEqual(self, first, second, error_message,
                     description='', traceback: str = None,
                     level: str = 'info'):
        self._assert((first == second),
                     error_message, description, traceback, level)

    def _assertIsValidHttpsUrl(self, check, error_message,
                               description='', traceback: str = None,
                               level: str = 'info'):
        self._assert(re.match('https://', check if check else ''),
                     description, traceback, level)

    def _assertIsValidHttpUrl(self, check, error_message,
                              description='', traceback: str = None,
                              level: str = 'info'):
        self._assert(re.match('https?://', check if check else ''),
                     description, traceback, level)

    # maybe useful .. one day ?!
        # idp_server = self.idp()
        # pysaml2 auth req object with signature check

        # SPID IdP not completely compliant to SAML2-core about Destination attribute
        # saml2.request:http://localhost:8088 not in ['http://localhost:54321/sso']
        # try:
            # req_obj = idp_server.parse_authn_request(self.authn_request_encoded,
            # BINDING_HTTP_POST)
        # except OtherError as e:
            # pass



import re

from spid_sp_test.utils import del_ns
from spid_sp_test import constants
from spid_sp_test import BASE_DIR, AbstractSpidCheck
import logging
import os
import requests
import xmlschema
import sys
import subprocess
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from lxml import etree
from tempfile import NamedTemporaryFile
from . metadata_public import SpidSpMetadataCheckPublic
from . metadata_private import SpidSpMetadataCheckPrivate
from . metadata_ag import SpidSpMetadataCheckAG
sys.path.append(os.path.join(os.path.dirname(__file__), os.pardir))


logger = logging.getLogger(__name__)


class SpidSpMetadataCheck(AbstractSpidCheck,
                          SpidSpMetadataCheckPublic,
                          SpidSpMetadataCheckPrivate,
                          SpidSpMetadataCheckAG):

    def __init__(self,
                 metadata_url,
                 xsds_files: list = None,
                 xsds_files_path: str = None,
                 production: bool = False):

        super(SpidSpMetadataCheck, self).__init__(verify_ssl=production)
        self.category = 'metadata_strict'

        self.logger = logger
        self.metadata_url = metadata_url
        self.production = production
        self.metadata = self.get(metadata_url)
        self.xsds_files_path = xsds_files_path or f'{BASE_DIR}/xsd'

        self.doc = etree.fromstring(self.metadata)
        # clean up namespace (otherwise xpath doesn't work ...)
        del_ns(self.doc)

    def get(self, metadata_url: str):
        if metadata_url[0:7] == 'file://':
            return open(metadata_url[7:], 'rb').read()
        else:
            request = requests.get(metadata_url,
                                   allow_redirects = True,
                                   verify=self.production)
            if request.status_code != 200:
                raise Exception(
                    f'Metadata not found: server response with code {request.status_code}'
                )
            else:
                return request.content

    def xsd_check(self,
                  xsds_files:list = ['saml-schema-metadata-2.0.xsd']):
        _msg = f'Found metadata'
        self.handle_result('debug',
                           _msg,
                           description = self.metadata.decode())
        _orig_pos = os.getcwd()
        os.chdir(self.xsds_files_path)
        metadata = self.metadata.decode()
        for testf in xsds_files:
            try:
                schema_file = open(testf, 'rb')
                msg = f'Test {self.metadata_url} with {schema_file.name}'
                schema = xmlschema.XMLSchema(schema_file)
                if not schema.is_valid(metadata):
                    schema.validate(metadata)
                    self.handle_result('error', ' '.join((msg)))
                    # raise Exception('Validation Error')
                logger.info(' '.join((msg, '-> OK')))
                break
            except Exception as e:
                os.chdir(_orig_pos)
                logger.error(f'{msg}: {e}')
                self.handle_result('error',
                                   msg,
                                   description='xsd test failed',
                                   traceback=f'{e}')
        os.chdir(_orig_pos)
        return self.is_ok(f'{self.__class__.__name__}.xsd_check')

    def test_metadata_no_newlines(self):
        self._assertFalse(
            re.match(r'^[\t\n\s\r\ ]*', self.metadata),
            (f'The XML of metadata should not '
              'contains newlines at the beginning.'),
             description = self.metadata[0:10],
             level='warning'
        )
        return self.is_ok(f'{self.__class__.__name__}.test_metadata_no_newlines')

    def test_EntityDescriptor(self):
        entity_desc = self.doc.xpath('//EntityDescriptor')
        desc = [etree.tostring(ent).decode() for ent in entity_desc if entity_desc]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertTrue(
            self.doc.attrib.get('entityID'),
            (f'Missing entityID in {self.doc.attrib}: '
             'The entityID attribute MUST be present - TR pag. 19'),
             description = self.doc.attrib.get('entityID'),
        )

        self._assertTrue(
            len(entity_desc) == 1,
            'Only one EntityDescriptor element MUST be present - TR pag. 19',
             description = self.doc.attrib.get('entityID'),
        )

        self._assertTrue(
            entity_desc[0].get('entityID'),
            'The entityID attribute MUST have a value - TR pag. 19',
             description = entity_desc[0].get('entityID'),
        )

        if self.production:
            self._assertIsValidHttpsUrl(
                self.doc.attrib.get('entityID'),
                'The entityID attribute MUST be a valid HTTPS url'
            )

        self._assertTrue((self.doc.attrib.get('entityID') == self.metadata_url),
                         f'The EntityID MUST be equal to {self.metadata_url}',
                         description=f"{self.doc.attrib.get('entityID')}",
                         level='warning')

        return self.is_ok(f'{self.__class__.__name__}.test_EntityDescriptor')

    def test_SPSSODescriptor(self):
        spsso = self.doc.xpath('//EntityDescriptor/SPSSODescriptor')
        desc = [etree.tostring(ent).decode() for ent in spsso if spsso]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertTrue((len(spsso) == 1),
                         'Only one SPSSODescriptor element MUST be present',
                         **error_kwargs)
        return self.is_ok(f'{self.__class__.__name__}.test_SPSSODescriptor')

    def test_SPSSODescriptor_SPID(self):
        spsso = self.doc.xpath('//EntityDescriptor/SPSSODescriptor')
        desc = [etree.tostring(ent).decode() for ent in spsso if spsso]
        error_kwargs = dict(description = desc) if desc else {}

        for attr in ['protocolSupportEnumeration', 'AuthnRequestsSigned']:
            self._assertTrue((attr in spsso[0].attrib),
                             f'The {attr} attribute MUST be present - TR pag. 20',
                             **error_kwargs)

            a = spsso[0].get(attr)
            self._assertIsNotNone(
                a,
                f'The {attr} attribute MUST have a value - TR pag. 20',
                **error_kwargs)

            if attr == 'AuthnRequestsSigned' and a:
                self._assertEqual(
                    a.lower(),
                    'true',
                    f'The {attr} attribute MUST be true - TR pag. 20',
                    **error_kwargs)

        return self.is_ok(
            f'{self.__class__.__name__}.test_SPSSODescriptor_SPID')


    def test_NameIDFormat_Transient(self):
        spsso = self.doc.xpath('//EntityDescriptor/SPSSODescriptor/NameIDFormat')
        desc = [etree.tostring(ent).decode() for ent in spsso if spsso]
        error_kwargs = dict(description = desc) if desc else {}

        if spsso:
            _rule = "urn:oasis:names:tc:SAML:2.0:nameid-format:transient"
            self._assertTrue(
                (spsso[0].text == _rule),
                f'The NameIDFormat MUST {_rule}',
                **error_kwargs)

        return self.is_ok(
            f'{self.__class__.__name__}.test_NameIDFormat_SPID')


    def test_xmldsig(self):
        '''Verify the SP metadata signature'''
        tmp_file = NamedTemporaryFile()
        tmp_file.write(self.metadata)
        tmp_file.seek(0)
        xmlsec_cmd = ['xmlsec1',
                      '--verify',
                      '--insecure',
                      '--id-attr:ID',
                      'urn:oasis:names:tc:SAML:2.0:metadata:'
                      'EntityDescriptor',
                      tmp_file.name]
        cmd = ' '.join(xmlsec_cmd)
        is_valid = True
        msg = 'the metadata signature MUST be valid - TR pag. 19'

        try:
            subprocess.run(cmd, shell=True, check=True,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as err:
            is_valid = False
            lines = [msg]
            if err.stderr:
                stderr = (
                    'stderr: ' +
                    '\nstderr: '.join(
                        list(
                            filter(
                                None,
                                err.stderr.decode('utf-8').split(r'\n')
                            )
                        )
                    )
                )
                lines.append(stderr)
            if err.stdout:
                stdout = (
                    'stdout: ' +
                    '\nstdout: '.join(
                        list(
                            filter(
                                None,
                                err.stdout.decode('utf-8').split(r'\n')
                            )
                        )
                    )
                )
                lines.append(stdout)
            _msg = '\n'.join(lines)
            self.handle_result('error', msg,
                               description='Description',
                               traceback=_msg)
            return

        xmlsec_cmd_string = ' '.join(xmlsec_cmd)
        _msg = f'{self.__class__.__name__}.test_xmldsig: OK'
        self.handle_result('info',
                           _msg, description=f"`{xmlsec_cmd_string}`")
        return is_valid

    def test_Signature(self):
        '''Test the compliance of Signature element'''
        sign = self.doc.xpath('//EntityDescriptor/Signature')
        desc = [etree.tostring(ent).decode() for ent in sign if sign]
        error_kwargs = dict(description = desc) if desc else {}
        self._assertTrue((len(sign) > 0),
                         'The Signature element MUST be present - TR pag. 19',
                         **error_kwargs)

        error_kwargs = dict(description = desc, traceback = '')
        if not sign:
            self.handle_result(
                'error',
                'The SignatureMethod element MUST be present - TR pag. 19',
                **error_kwargs
            )
            self.handle_result(
                'error',
                'The Algorithm attribute MUST be present in SignatureMethod element - TR pag. 19',
                **error_kwargs
            )
            self.handle_result(
                'error',
                "The signature algorithm MUST be valid - TR pag. 19",
                description = f"Must be one of [{', '.join(constants.ALLOWED_XMLDSIG_ALGS)}]"
            )
            self.handle_result(
                'error',
                'The Algorithm attribute MUST be present in DigestMethod element - TR pag. 19',
                **error_kwargs
            )
            self.handle_result(
                'error',
                f"The digest algorithm MUST be valid - TR pag. 19",
                description = f"Must be one of [{', '.join(constants.ALLOWED_DGST_ALGS)}]"
            )
        else:
            method = sign[0].xpath('./SignedInfo/SignatureMethod')
            desc = [etree.tostring(ent).decode() for ent in method if method]
            error_kwargs = dict(description = desc) if desc else {}
            self._assertTrue((len(method) > 0),
                             'The SignatureMethod element MUST be present - TR pag. 19',
                             **error_kwargs)

            self._assertTrue(('Algorithm' in method[0].attrib),
                             'The Algorithm attribute MUST be present '
                             'in SignatureMethod element - TR pag. 19',
                             **error_kwargs)

            alg = method[0].get('Algorithm')
            self._assertIn(alg, constants.ALLOWED_XMLDSIG_ALGS,
                           'The signature algorithm MUST be valid - TR pag. 19',
                           description = f"One of {(', '.join(constants.ALLOWED_XMLDSIG_ALGS))}"
            )

            method = sign[0].xpath('./SignedInfo/Reference/DigestMethod')
            self._assertTrue((len(method) == 1),
                             'The DigestMethod element MUST be present - TR pag. 19',
                             **error_kwargs)

            self._assertTrue(('Algorithm' in method[0].attrib),
                             'The Algorithm attribute MUST be present '
                             'in DigestMethod element - TR pag. 19',
                             **error_kwargs)

            alg = method[0].get('Algorithm')
            self._assertIn(alg, constants.ALLOWED_DGST_ALGS,
                           'The digest algorithm MUST be valid - TR pag. 19',
                           description = f"One of {(', '.join(constants.ALLOWED_DGST_ALGS))}",
            )


        return self.is_ok(f'{self.__class__.__name__}.test_Signature')

    def test_KeyDescriptor(self):
        '''Test the compliance of KeyDescriptor element(s)'''
        kds = self.doc.xpath('//EntityDescriptor/SPSSODescriptor'
                             '/KeyDescriptor[@use="signing"]')
        self._assertGreaterEqual(
            len(kds), 1,
            'At least one signing KeyDescriptor MUST be present - TR pag. 19'
        )

        desc = [etree.tostring(ent).decode() for ent in kds if kds]
        error_kwargs = dict(description = desc, traceback = '')

        for kd in kds:
            certs = kd.xpath('./KeyInfo/X509Data/X509Certificate')
            self._assertGreaterEqual(
                len(certs), 1,
                'At least one signing x509 MUST be present - TR pag. 19',
                **error_kwargs)

        kds = self.doc.xpath('//EntityDescriptor/SPSSODescriptor'
                             '/KeyDescriptor[@use="encryption"]')

        for kd in kds:
            certs = kd.xpath('./KeyInfo/X509Data/X509Certificate')
            self._assertGreaterEqual(
                len(certs), 1,
                'At least one encryption x509 MUST be present - TR pag. 19',
                **error_kwargs)

        return self.is_ok(f'{self.__class__.__name__}.test_KeyDescriptor')

    def test_SingleLogoutService(self):
        '''Test the compliance of SingleLogoutService element(s)'''
        slos = self.doc.xpath('//EntityDescriptor/SPSSODescriptor'
                              '/SingleLogoutService')
        self._assertGreaterEqual(
            len(slos),
            1,
            'One or more SingleLogoutService elements MUST be present - AV n. 3'
        )

        desc = [etree.tostring(ent).decode() for ent in slos if slos]
        error_kwargs = dict(description = desc)

        for slo in slos:
            for attr in ['Binding', 'Location']:
                self._assertTrue(
                    (attr in slo.attrib),
                    f'The {attr} attribute in SingleLogoutService element MUST be present - AV n. 3',
                    **error_kwargs)

                a = slo.get(attr)
                self._assertIsNotNone(
                    a,
                    f'The {attr} attribute in SingleLogoutService element MUST have a value',
                    **error_kwargs
                )

                if attr == 'Binding':
                    self._assertIn(
                        a,
                        constants.ALLOWED_SINGLELOGOUT_BINDINGS,
                        (('The %s attribute in SingleLogoutService element MUST be one of [%s] - AV n. 3') %  # noqa
                         (attr, ', '.join(constants.ALLOWED_BINDINGS))),
                        **error_kwargs # noqa
                    )
                if attr == 'Location' and self.production:
                    self._assertIsValidHttpsUrl(
                        a,
                        f'The {attr} attribute '
                        'in SingleLogoutService element '
                        'MUST be a valid HTTPS URL - AV n. 1 and n. 3',
                        **error_kwargs
                    )
        return self.is_ok(f'{self.__class__.__name__}.test_SingleLogoutService')

    def test_AssertionConsumerService(self):
        '''Test the compliance of AssertionConsumerService element(s)'''
        acss = self.doc.xpath('//EntityDescriptor/SPSSODescriptor'
                              '/AssertionConsumerService')

        desc = [etree.tostring(ent).decode() for ent in acss if acss]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertGreaterEqual(len(acss), 1,
                                 'At least one AssertionConsumerService '
                                 'MUST be present - TR pag. 20')

        for acs in acss:
            for attr in ['index', 'Binding', 'Location']:
                self._assertTrue(
                    (attr in acs.attrib),
                    f'The {attr} attribute MUST be present - TR pag. 20'
                )
                a = acs.get(attr)
                if attr == 'index':
                    self._assertGreaterEqual(
                        int(a),
                        0,
                        f'The {attr} attribute MUST be >= 0 - TR pag. 20',
                        **error_kwargs
                    )
                elif attr == 'Binding':
                    self._assertIn(
                        a, constants.ALLOWED_BINDINGS,
                        (('The %s attribute MUST be one of [%s] - TR pag. 20') %
                         (attr, ', '.join(constants.ALLOWED_BINDINGS))),
                        **error_kwargs
                    )
                elif attr == 'Location' and self.production:
                    self._assertIsValidHttpsUrl(
                        a,
                        f'The {attr} attribute MUST be a '
                        'valid HTTPS url - TR pag. 20 and AV n. 1',
                        **error_kwargs
                    )
                else:
                    pass

        return self.is_ok(f'{self.__class__.__name__}.test_AssertionConsumerService')

    def test_AssertionConsumerService_SPID(self):
        acss = self.doc.xpath('//EntityDescriptor/SPSSODescriptor'
                              '/AssertionConsumerService'
                              '[@isDefault="true"]')
        desc = [etree.tostring(ent).decode() for ent in acss if acss]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertTrue((len(acss) == 1),
                         'Only one default AssertionConsumerService '
                         'MUST be present - TR pag. 20',
                         **error_kwargs)

        acss = self.doc.xpath('//EntityDescriptor/SPSSODescriptor'
                              '/AssertionConsumerService'
                              '[@index="0"]'
                              '[@isDefault="true"]')
        self._assertTrue((len(acss) == 1),
                         'Must be present the default AssertionConsumerService '
                         'with index = 0 - TR pag. 20',
                         **error_kwargs)
        return self.is_ok(
            f'{self.__class__.__name__}.test_AssertionConsumerService_SPID')

    def test_AttributeConsumingService(self):
        '''Test the compliance of AttributeConsumingService element(s)'''
        acss = self.doc.xpath('//EntityDescriptor/SPSSODescriptor'
                              '/AttributeConsumingService')

        desc = [etree.tostring(ent).decode() for ent in acss if acss]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertGreaterEqual(
            len(acss),
            1,
            'One or more AttributeConsumingService elements MUST be present - TR pag. 20',
            **error_kwargs
        )
        return self.is_ok(f'{self.__class__.__name__}.test_AttributeConsumingService')


    def test_AttributeConsumingService_SPID(self):
        acss = self.doc.xpath('//EntityDescriptor/SPSSODescriptor'
                              '/AttributeConsumingService')

        desc = [etree.tostring(ent).decode() for ent in acss if acss]
        error_kwargs = dict(description = desc) if desc else {}
        for acs in acss:
            self._assertTrue(
                ('index' in acs.attrib),
                'The index attribute in AttributeConsumigService element MUST be present',
                **error_kwargs)

            idx = int(acs.get('index'))
            self._assertGreaterEqual(
                idx,
                0,
                'The index attribute in AttributeConsumigService '
                'element MUST be >= 0 - TR pag. 20',
                **error_kwargs
            )

            sn = acs.xpath('./ServiceName')
            self._assertTrue((len(sn) > 0),
                             'The ServiceName element MUST be present',
                             **error_kwargs)
            for sns in sn:
                self._assertIsNotNone(
                    sns.text,
                    'The ServiceName element MUST have a value',
                    **error_kwargs)

            ras = acs.xpath('./RequestedAttribute')
            self._assertGreaterEqual(
                len(ras),
                1,
                'One or more RequestedAttribute elements MUST be present - TR pag. 20',
                **error_kwargs
            )

            for ra in ras:
                self._assertTrue(
                    ('Name' in ra.attrib),
                    'The Name attribute in RequestedAttribute element '
                    'MUST be present - TR pag. 20 and AV n. 6',
                    **error_kwargs)

                self._assertIn(
                    ra.get('Name'),
                    constants.SPID_ATTRIBUTES,
                    f'The "{ra.attrib.values()[0]}" attribute in RequestedAttribute element MUST be valid',
                    description = f"one of [{', '.join(constants.SPID_ATTRIBUTES)}] - TR pag. 20 and AV n.6"
                )

            al = acs.xpath('RequestedAttribute/@Name')
            self._assertEqual(
                len(al),
                len(set(al)),
                'AttributeConsumigService MUST not contain duplicated RequestedAttribute - TR pag. 20',
                **error_kwargs
            )
        return self.is_ok(f'{self.__class__.__name__}.test_AttributeConsumingService_SPID')

    def test_Organization(self):
        '''Test the compliance of Organization element'''
        orgs = self.doc.xpath('//EntityDescriptor/Organization')

        desc = [etree.tostring(ent).decode() for ent in orgs if orgs]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertTrue(
            (len(orgs) == 1),
            'Only one Organization element can be present - TR pag. 20'
        )

        enames = ['OrganizationName',
                  'OrganizationDisplayName',
                  'OrganizationURL']
        lang_counter = dict()

        if len(orgs) == 1:
            org = orgs[0]
            for ename in enames:
                elements = org.xpath(f'./{ename}')
                self._assertGreater(
                    len(elements),
                    0,
                    f'One or more {ename} elements MUST be present - TR pag. 20',
                    **error_kwargs
                )

                for element in elements:
                    self._assertTrue(
                        ('{http://www.w3.org/XML/1998/namespace}lang' in element.attrib),  # noqa
                        f'The lang attribute in {ename} element MUST be present - TR pag. 20', # noqa
                        **error_kwargs
                    )

                    lang = element.attrib.items()[0][1]
                    if lang_counter.get(lang):
                        lang_counter[lang] += 1
                    else:
                        lang_counter[lang] = 1


                    self._assertIsNotNone(
                        element.text,
                        f'The {ename} element MUST have a value - TR pag. 20',
                        **error_kwargs
                    )

                    if ename == 'OrganizationURL' and self.production:
                        OrganizationURLvalue = element.text.strip()
                        if not (OrganizationURLvalue.startswith('http://') or
                                OrganizationURLvalue.startswith('https://')):
                            OrganizationURLvalue = f'https://{OrganizationURLvalue}'
                        self._assertIsValidHttpUrl(
                            OrganizationURLvalue,
                            f'The {ename} -element MUST be a valid URL - TR pag. 20',
                            **error_kwargs
                        )

            # lang counter check
            for k,v in lang_counter.items():
                num_enames = len(enames)
                self._assertTrue(
                        (v == num_enames),
                        (f'The elements OrganizationName, OrganizationDisplayName and OrganizationURL '
                         'MUST have the same number of lang attributes'), # noqa
                        **error_kwargs
                    )

            self._assertTrue(
                    ("it" in lang_counter),
                    (f'The elements OrganizationName, OrganizationDisplayName and OrganizationURL '
                     'MUST have at least an it language enabled'), # noqa
                    **error_kwargs
                )

        return self.is_ok(f'{self.__class__.__name__}.test_Organization')

    def test_profile_saml2core( self):
        self.xsd_check(
            xsds_files = ['saml-schema-metadata-2.0.xsd']
        )

        # loop for all the attrs that starts with test_ ... todo?
        self.test_EntityDescriptor()

        self.test_SPSSODescriptor()
        self.test_NameIDFormat_Transient()
        self.test_xmldsig()
        self.test_Signature()
        self.test_KeyDescriptor()
        self.test_SingleLogoutService()
        self.test_AssertionConsumerService()
        self.test_AttributeConsumingService()
        self.test_Organization()

    def test_profile_spid_sp(self):
        self.test_profile_saml2core()

        self.xsd_check(
            xsds_files = ['saml-schema-metadata-sp-spid.xsd',
                          'saml-schema-metadata-sp-spid-av29.xsd']
        )
        self.test_SPSSODescriptor_SPID()
        self.test_AssertionConsumerService_SPID()
        self.test_AttributeConsumingService_SPID()
        self.test_contactperson_email()
        self.test_contactperson_phone()

    def test_profile_spid_sp_public(self):
        self.test_profile_spid_sp()
        self.test_Contacts_PubPriv()
        self.test_Extensions_PubPriv()
        self.test_Contacts_VATFC()
        self.test_Contacts_IPACode()
        self.test_extensions_public_private(ext_type="Public")

    def test_profile_spid_sp_private(self):
        self.test_profile_spid_sp()
        self.test_Contacts_PubPriv()
        self.test_Contacts_PubPriv(contact_type='billing')
        self.test_Extensions_PubPriv()
        self.test_extensions_public_private(ext_type="Private")

        # invalid ! to be removed soon
        # self.test_contactperson_email(
            # email_xpath="//ContactPerson/Extensions/CessionarioCommittente/EmailAddress"
        # )

        self.test_Contacts_VATFC(private=True)
        self.test_Contacts_Priv()
        self.xsd_check(xsds_files = [
            'saml-schema-metadata-2.0.xsd',
            'spid-invoicing.xsd'
        ])

    def test_profile_spid_ag_public_full(self):
        self.test_profile_spid_sp()

        self.test_extensions_public_private(ext_type="Public")
        self.test_Contacts_IPACode()
        self.test_Contacts_VATFC()
        self.test_extensions_public_ag()
        self.test_Extensions_PubPriv()

        # The ContactPerson element of contactType “other” and spid:entityType “spid:aggregator” MUST be present
        # The ContactPerson element of contactType “other” and spid:entityType “spid:aggregated” MUST be present
        self.test_Contacts_PubPriv(contact_type="aggregator")
        self.test_Contacts_PubPriv(contact_type="aggregated")

        # The entityID MUST not contains the query-string part
        self.test_entityid_qs()

        # The entityID MUST contain the activity code “pub-ag-full”
        self.test_entityid_contains(value='pub-ag-full')

        # The PublicServicesFullAggregator element MUST be present
        self.test_extensions_public_ag(
            ext_types = ["//ContactPerson/Extensions/PublicServicesFullAggregator"],
            must=True
        )


    def test_profile_spid_ag_public_lite(self):
        self.test_profile_spid_sp()
        self.test_extensions_public_private(ext_type="Public")

        # The entityID MUST contain the activity code “pub-ag-lite”
        self.test_entityid_contains(value='pub-ag-lite')

        # Only one ContactPerson element of contactType “other” and spid:entityType “spid:aggregator” MUST be present
        # Only one ContactPerson element of contactType “other” and spid:entityType “spid:aggregated” MUST be present
        self.test_Contacts_PubPriv(contact_type="aggregator")
        self.test_Contacts_PubPriv(contact_type="aggregated")

        # TODO
        # If the ContactPerson is of spid:entityType “spid:aggregator” the Extensions element MUST contain the element spid:KeyDescriptor with attribute use “spid:validation”

        # The PublicServicesLightAggregator element MUST be present
        self.test_extensions_public_ag(
            ext_types = ["//ContactPerson/Extensions/PublicServicesLightAggregator"],
            must=True
        )

    def test_profile_spid_op_public_full(self):
        self.test_profile_spid_sp()

        self.test_Contacts_VATFC()
        self.test_extensions_public_private(ext_type="Public")

        # The entityID MUST contain the activity code “pub-op-full”
        self.test_entityid_contains(value='pub-ag-full')

        # Only one ContactPerson element of contactType “other” and spid:entityType “spid:aggregator” MUST be present
        self.test_Contacts_PubPriv(contact_type="aggregator")

        # The PublicServicesFullOperator element MUST be present
        self.test_extensions_public_ag(
            ext_types = ["//ContactPerson/Extensions/PublicServicesFullOperator"],
            must=True
        )


    def test_profile_spid_op_public_lite(self):
        self.test_profile_spid_sp()

        self.test_Contacts_VATFC()
        self.test_extensions_public_private(ext_type="Public")

        # The entityID MUST contain the activity code “pub-op-lite”
        self.test_entityid_contains(value='pub-op-lite')

        # Only one ContactPerson element of contactType “other” and spid:entityType “spid:aggregator” MUST be present
        # Only one ContactPerson element of contactType “other” and spid:entityType “spid:aggregated” MUST be present
        self.test_Contacts_PubPriv(contact_type="aggregator")
        self.test_Contacts_PubPriv(contact_type="aggregated")

        # The PublicServicesLightOperator element MUST be present
        self.test_extensions_public_ag(
            ext_types = ["//ContactPerson/Extensions/PublicServicesLightOperator"],
            must=True
        )



import re

import base64
import copy
import logging
import os
import requests
import xmlschema
import sys
import subprocess
import urllib
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from lxml import etree
from spid_sp_test.utils import (del_ns,
                                samlreq_from_htmlform,
                                decode_authn_req_http_redirect)
from spid_sp_test.idp.settings import SAML2_IDP_CONFIG
from spid_sp_test import constants
from spid_sp_test import BASE_DIR, AbstractSpidCheck

from saml2.server import Server
from saml2.sigver import CryptoBackendXMLSecurity
# from saml2.sigver import CryptoBackendXmlSec1
sys.path.append(os.path.join(os.path.dirname(__file__), os.pardir))
from tempfile import NamedTemporaryFile

from . exceptions import SAMLRequestNotFound


logger = logging.getLogger(__name__)


def get_authn_request(authn_request_url, verify_ssl=False):
    """
        Detects the auth request url, if http/xml file or html file
    """
    data = {}
    request = None
    binding = 'post' or 'redirect'
    authn_request_str = None
    requests_session = None

    if authn_request_url[0:7] == 'file://':
        authn_request = open(authn_request_url[7:], 'rb').read().strip().strip(b'\n')
        # stupid test ... good enough for now
        authn_request_str = authn_request.decode()
        if authn_request_str[0] == '<' and authn_request_str[-1] == '>':
            binding = 'post'
        elif '?' in authn_request_str and '&' in authn_request_str:
            binding = 'redirect'
        else:
            raise Exception(f"Can't detect authn request from f{authn_request_url}")
    else:
        requests_session = requests.Session()
        request = requests_session.get(
            authn_request_url,
            verify=verify_ssl,
            allow_redirects=False
        )
        if request.status_code not in (200, 302):
            raise Exception(('Authn Request page returns a HTML error '
                             f'code: {request.status_code}'))
        elif request.headers.get('Location'):
            binding = 'redirect'
        else:
            binding = 'post'

    if binding == 'redirect':
        # HTTP-REDIRECT
        redirect = request.headers['Location'] if request else authn_request_str
        q_args = urllib.parse.splitquery(redirect)[1]
        authn_request = dict(urllib.parse.parse_qsl(q_args))

        if not authn_request.get('SAMLRequest'):
            logger.critical('\nHTTP-REDIRECT without any SAMLRequest in. '
                            'Is this SP behind a Proxy or is there any '
                            f'DiscoveryService enabled? {authn_request}')
            sys.exit(1)

        data['SAMLRequest_redirect'] = redirect
        data['SAMLRequest'] = authn_request['SAMLRequest']
        data['SAMLRequest_xml'] = decode_authn_req_http_redirect(
            authn_request['SAMLRequest']).encode()
        data['RelayState'] = authn_request.get('RelayState')
        data['SigAlg'] = authn_request['SigAlg']
        data['Signature'] = authn_request['Signature']

    elif binding == 'post':
        # HTTP POST
        authn_request_str = request.content.decode() if request else authn_request_str
        form_dict = samlreq_from_htmlform(authn_request_str)
        if form_dict:
            data['action'] = form_dict['action']
            data['method'] = form_dict['method']
            data['SAMLRequest'] = form_dict['SAMLRequest']
            data['SAMLRequest_xml'] = base64.b64decode(
                form_dict['SAMLRequest'].encode())
            data['RelayState'] = form_dict['RelayState']
        elif ':AuthnRequest ' in authn_request_str:
            data = {'SAMLRequest_xml': authn_request_str.encode(),
                    'SAMLRequest' : base64.b64encode(authn_request),
                    'RelayState' : '/'
                    }
        else:
            raise SAMLRequestNotFound(f'{authn_request_str}')
    else:
        raise SAMLRequestNotFound()

    if requests_session:
        data['requests_session'] = requests_session
    return data


class SpidSpAuthnReqCheck(AbstractSpidCheck):
    xsds_files = [
        'saml-schema-protocol-2.0.xsd',
    ]

    def __init__(self,
                 metadata,
                 authn_request_url: str = None,
                 authn_request: dict = {},
                 xsds_files: list = None,
                 xsds_files_path: str = None,
                 production: bool = False):

        super(SpidSpAuthnReqCheck, self).__init__(verify_ssl=production)
        self.category = 'authnrequest_strict'

        self.logger = logger
        self.metadata = metadata

        self.authn_request = get_authn_request(authn_request_url,
                                               verify_ssl=production)

        try:
            self.authn_request_decoded = self.authn_request['SAMLRequest_xml']
            self.authn_request_encoded = self.authn_request['SAMLRequest']

        except KeyError:
            raise SAMLRequestNotFound(self.authn_request)

        self.relay_state = self.authn_request.get('RelayState') or ""

        self.xsds_files = xsds_files or self.xsds_files
        self.xsds_files_path = xsds_files_path or f'{BASE_DIR}/xsd'

        self.md = etree.fromstring(self.metadata)
        del_ns(self.md)

        self.doc = etree.fromstring(self.authn_request_decoded)
        # clean up namespace (otherwise xpath doesn't work ...)
        del_ns(self.doc)

        # binding detection
        self.IS_HTTP_REDIRECT = self.authn_request.get('Signature')
        # HTTP-REDIRECT params
        self.params = {'RelayState': self.relay_state}
        self.production = production

    def idp(self):
        idp_config = copy.deepcopy(SAML2_IDP_CONFIG)
        idp_server = Server(idp_config)
        if self.metadata:
            idp_server.metadata.imp(
                [
                    {"class": "saml2.mdstore.InMemoryMetaData",
                     "metadata": [(self.metadata,)]
                     }
                ]
            )
        return idp_server

    def test_xsd(self):
        '''Test if the XSD validates and if the signature is valid'''

        _orig_pos = os.getcwd()
        os.chdir(self.xsds_files_path)
        authn_request = self.authn_request_decoded.decode()
        schema_file = open('saml-schema-protocol-2.0.xsd', 'rb')
        msg = f'Test authn_request with {schema_file.name}'
        try:
            schema = xmlschema.XMLSchema(schema_file)
            if not schema.is_valid(authn_request):
                schema.validate(authn_request)
                self.handle_result('error',
                                   ' '.join((msg, '-> FAILED!')))
                raise Exception('Validation Error')
            logger.info(' '.join((msg, '-> OK')))
        except Exception as e:
            os.chdir(_orig_pos)
            self.handle_result('error', '-> '.join((msg, f'{e}')))
        os.chdir(_orig_pos)

        return self.is_ok(f'{self.__class__.__name__}.test_xsd')

    def test_xmldsig(self):
        certs = self.md.xpath(
            '//SPSSODescriptor/KeyDescriptor[@use="signing"]'
            '/KeyInfo/X509Data/X509Certificate/text()')

        desc = certs
        error_kwargs = dict(description = desc) if desc else {}

        msg = ('The AuthnRequest MUST validate against XSD '
               'and MUST have a valid signature')

        if not certs:
            self.handle_result('error', '-> '.join(
                (msg, 'AuthnRequest Signature validation failed: certificates are missing.'),
                **error_kwargs)
            )
            return self.is_ok(f'{self.__class__.__name__}.test_xsd_and_xmldsig')
        else:
            is_valid = False
            for cert in certs:
                if self.IS_HTTP_REDIRECT:
                    with NamedTemporaryFile(suffix='.xml') as cert_file:
                        if cert[-1] != '\n':
                            cert += '\n'
                        cert_file.write(
                            f'-----BEGIN CERTIFICATE-----\n{cert}-----END CERTIFICATE-----'.encode()
                        )
                        cert_file.seek(0)
                        _sigalg = self.authn_request.get('SigAlg', "")
                        quoted_req = urllib.parse.quote_plus(self.authn_request['SAMLRequest'])
                        quoted_rs = urllib.parse.quote_plus(self.authn_request.get('RelayState') or "")
                        quoted_sigalg = urllib.parse.quote_plus(_sigalg)
                        authn_req = (f"SAMLRequest={quoted_req}&"
                                     f"RelayState={quoted_rs}&"
                                     f"SigAlg={quoted_sigalg}")

                        payload_file = NamedTemporaryFile(suffix='.xml')
                        payload_file.write(authn_req.encode())
                        payload_file.seek(0)

                        signature_file = NamedTemporaryFile(suffix='.sign')
                        signature_file.write(base64.b64decode(self.authn_request['Signature'].encode()))
                        signature_file.seek(0)

                        pubkey_file = NamedTemporaryFile(suffix='.crt')
                        x509_cert = subprocess.getoutput(
                            f'openssl x509 -in {cert_file.name} -noout -pubkey'
                        )
                        pubkey_file.write(x509_cert.encode())
                        pubkey_file.seek(0)

                        dgst = _sigalg.split('-')[-1]
                        signature = signature_file.name

                        ver_cmd = (f'openssl dgst -{dgst} '
                                   f'-verify {pubkey_file.name} '
                                   f'-signature {signature} {payload_file.name}')
                        exit_msg = subprocess.getoutput(ver_cmd)
                        error_kwargs['description'] = exit_msg
                        if 'Verified OK' in exit_msg:
                            is_valid = True
                        else:
                            is_valid = False

                else:
                    # pyXMLSecurity allows to pass a certificate without store it on a file
                    backend = CryptoBackendXMLSecurity()
                    is_valid = backend.validate_signature(self.authn_request_decoded,
                                                          cert_file=cert,
                                                          cert_type='pem',
                                                          node_name=constants.NODE_NAME,
                                                          node_id=None)
                if is_valid:
                    break

        self._assertTrue(is_valid,
                         'AuthnRequest Signature validation failed',
                         **error_kwargs)
        return self.is_ok(f'{self.__class__.__name__}.test_xmldsig')

    def test_authnrequest_no_newlines(self):
        self._assertFalse(
            re.match(r'^[\t\n\s\r\ ]*', self.authn_request_decoded),
            (f'The XML of authn request should not '
              'contains newlines at the beginning.'),
             description = self.metadata[0:10],
             level='warning'
        )
        return self.is_ok(f'{self.__class__.__name__}.test_authnrequest_no_newlines')

    def test_AuthnRequest(self):
        '''Test the compliance of AuthnRequest element'''
        req = self.doc.xpath('/AuthnRequest')

        req_desc = [etree.tostring(ent).decode() for ent in req if req]
        error_kwargs = dict(description = req_desc) if req_desc else {}

        self._assertTrue(
            (len(req) == 1),
            'One AuthnRequest element MUST be present',
            **error_kwargs
        )
        if req:
            req = req[0]
        else:
            return self.is_ok(f'{self.__class__.__name__}.test_AuthnRequest')

        for attr in ('ID', 'Version', 'IssueInstant', 'Destination'):
            self._assertTrue(
                (attr in req.attrib),
                f'The {attr} attribute MUST be present - TR pag. 8 ',
                **error_kwargs
            )

            value = req.get(attr)
            if (attr == 'ID'):
                self._assertIsNotNone(
                    value,
                    f'The {attr} attribute MUST have a value - TR pag. 8 ',
                    **error_kwargs
                )

            if (attr == 'Version'):
                exp = '2.0'
                self._assertEqual(
                    value,
                    exp,
                    f'The {attr} attribute MUST be {exp} - TR pag. 8 ',
                    **error_kwargs
                )

            if (attr == 'IssueInstant'):
                self._assertIsNotNone(
                    value,
                    f'The {attr} attribute MUST have a value - TR pag. 8 ',
                    **error_kwargs
                )
                self._assertTrue(
                    bool(constants.UTC_STRING.search(value)),
                    f'The {attr} attribute MUST be a valid UTC string - TR pag. 8 ',
                    **error_kwargs
                )

            if (attr == 'Destination'):
                self._assertIsNotNone(
                    value,
                    f'The {attr} attribute MUST have a value - TR pag. 8 ',
                    **error_kwargs
                )
                if self.production:
                    self._assertIsValidHttpsUrl(
                        value,
                        f'The {attr} attribute MUST be a valid HTTPS url - TR pag. 8 ',
                        **error_kwargs
                    )

        self._assertTrue(
            ('IsPassive' not in req.attrib),
            'The IsPassive attribute MUST not be present - TR pag. 9 ',
            **error_kwargs
        )
        return self.is_ok(f'{self.__class__.__name__}.test_AuthnRequest')

    def test_AuthnRequest_SPID(self):
        '''Test the compliance of AuthnRequest element'''
        req = self.doc.xpath('/AuthnRequest')[0]

        req_desc = [etree.tostring(ent).decode() for ent in req if req is not None]
        error_kwargs = dict(description = req_desc) if req_desc else {}

        acr = req.xpath('//RequestedAuthnContext/AuthnContextClassRef')
        acr_desc = [etree.tostring(_acr).decode() for _acr in acr]

        if acr:
            level = acr[0].text
            if bool(constants.SPID_LEVEL_23.search(level)):
                self._assertTrue(
                    ('ForceAuthn' in req.attrib),
                    'The ForceAuthn attribute MUST be present if SPID level > 1 - TR pag. 8 ',
                    description = acr_desc
                )
                value = req.get('ForceAuthn')
                if value:
                    self._assertTrue(
                        (value.lower() in constants.BOOLEAN_TRUE),
                        'The ForceAuthn attribute MUST be true or 1 - TR pag. 8 ',
                        **error_kwargs
                    )

        attr = 'AssertionConsumerServiceIndex'
        acss = self.md.xpath('//EntityDescriptor/SPSSODescriptor'
                             '/AssertionConsumerService')
        acss_desc = [etree.tostring(_acss).decode() for _acss in acss]

        if attr in req.attrib:
            value = req.get(attr)
            availableassertionindexes = []

            for acs in acss:
                index = acs.get('index')
                availableassertionindexes.append(index)

            self._assertTrue(
                value in availableassertionindexes,
                f'The {attr} attribute MUST be equal to an AssertionConsumerService index - TR pag. 8 ',
                description = acss_desc
            )

            self._assertIsNotNone(
                value,
                f'The {attr} attribute MUST have a value- TR pag. 8 ',
                **error_kwargs
            )
            self._assertGreaterEqual(
                int(value),
                0,
                f'The {attr} attribute MUST be >= 0 - TR pag. 8 and pag. 20',
                **error_kwargs
            )

        else:
            availableassertionlocations = []

            for acs in acss:
                location = acs.get('Location')
                availableassertionlocations.append(location)

            for attr in ['AssertionConsumerServiceURL', 'ProtocolBinding']:
                self._assertTrue(
                    (attr in req.attrib),
                    f'The {attr} attribute MUST be present - TR pag. 8 ',
                    **error_kwargs
                )

                value = req.get(attr)
                self._assertIsNotNone(
                    value,
                    f'The {attr} attribute MUST have a value - TR pag. 8 ',
                    **error_kwargs
                )

                if attr == 'AssertionConsumerServiceURL':
                    if self.production:
                        self._assertIsValidHttpsUrl(
                            value,
                            f'The {attr} attribute MUST be a valid HTTPS url - TR pag. 8 and pag. 16',
                            **error_kwargs
                        )

                    self._assertTrue(
                        value in availableassertionlocations,
                        f'The {attr} attribute MUST be equal to an AssertionConsumerService Location - TR pag. 8 ',
                        **error_kwargs
                    )

                if attr == 'ProtocolBinding':
                    exp = 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST'
                    self._assertEqual(
                        value,
                        exp,
                        f'The {attr} attribute MUST be {exp} - TR pag. 8 ',
                        **error_kwargs
                    )

        attr = 'AttributeConsumingServiceIndex'
        if attr in req.attrib:
            availableattributeindexes = []

            acss = self.md.xpath('//EntityDescriptor/SPSSODescriptor'
                                 '/AttributeConsumingService')
            for acs in acss:
                index = acs.get('index')
                availableattributeindexes.append(index)

            value = req.get(attr)
            self._assertIsNotNone(
                value,
                f'The {attr} attribute MUST have a value - TR pag. 8',
                **error_kwargs
            )
            self._assertGreaterEqual(
                int(value),
                0,
                f'The {attr} attribute MUST be >= 0 - TR pag. 8 and pag. 20',
                **error_kwargs
            )
            self._assertTrue(
                value in availableattributeindexes,
                f'The {attr} attribute MUST be equal to an AttributeConsumingService index - TR pag. 8 ',
                **error_kwargs
            )
        return self.is_ok(f'{self.__class__.__name__}.test_AuthnRequest_SPID')

    def test_AuthnRequest_SPID_extra(self):
        '''Test the compliance of AuthnRequest element'''

        # ForceAuthn MUST be true if 'Comparison' is 'minimum' and
        # SPID level is L1

        req = self.doc.xpath('/AuthnRequest')
        rac = None
        acr = None
        if req:
            rac = req[0].xpath('./RequestedAuthnContext')
        if rac:
            acr = rac[0].xpath('./AuthnContextClassRef')

        if req and rac and acr:
            req = req[0]
            rac = rac[0]
            acr = acr[0]

            if (rac.get('Comparison') in ('minimum', 'exact')
                    and acr.text in
                    ('https://www.spid.gov.it/SpidL2',
                     'https://www.spid.gov.it/SpidL3')):
                self._assertTrue(
                    ('ForceAuthn' in req.attrib),
                     'The ForceAuthn attribute MUST be present '
                     'because of minimum/SpidL2',
                     description = req.attrib
                )
                self._assertTrue(
                    req.get('ForceAuthn', "").lower() in ('true', 1, '1'),
                    'The ForceAuthn attribute MUST be True '
                    'because of minimum/SpidL2',
                    description = req.attrib
                )
        else:
            self.handle_error(
                'AuthnRequest or RequestAuthnContext or AuthnContextClassRef missing'
            )

        return self.is_ok(f'{self.__class__.__name__}.test_AuthnRequest_extra')

    def test_Subject(self):
        '''Test the compliance of Subject element'''

        subj = self.doc.xpath('//AuthnRequest/Subject')

        desc = [etree.tostring(ent).decode() for ent in subj if subj]
        error_kwargs = dict(description = desc) if desc else {}

        if len(subj) > 1:
            self._assertEqual(
                len(subj),
                1,
                'Only one Subject element can be present - TR pag. 9',
                **error_kwargs
            )

        if len(subj) == 1:
            subj = subj[0]
            name_id = subj.xpath('./NameID')
            self._assertEqual(
                len(name_id),
                1,
                'One NameID element in Subject element MUST be present - TR pag. 9',
                **error_kwargs
            )
            name_id = name_id[0]
            for attr in ['Format', 'NameQualifier']:
                self._assertTrue(
                    (attr in name_id.attrib),
                    f'The {attr} attribute MUST be present - TR pag. 9',
                    **error_kwargs
                )

                value = name_id.get(attr)

                self._assertIsNotNone(
                    value,
                    f'The {attr} attribute MUST have a value - TR pag. 9',
                    **error_kwargs
                )

                if attr == 'Format':
                    exp = 'urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified'
                    self._assertEqual(
                        value,
                        exp,
                        f'The {attr} attribute MUST be {exp} - TR pag. 9',
                        **error_kwargs
                    )
        return self.is_ok(f'{self.__class__.__name__}.test_Subject')

    def test_Issuer(self):
        '''Test the compliance of Issuer element'''
        e = self.doc.xpath('//AuthnRequest/Issuer')

        desc = [etree.tostring(ent).decode() for ent in e if e]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertTrue(
            (len(e) == 1),
            'One Issuer element MUST be present - TR pag. 9',
            error_kwargs
        )

        if not e:
            return self.is_ok(f'{self.__class__.__name__}.test_AuthnRequest')
        else:
            e = e[0]

        self._assertIsNotNone(
            e.text,
            'The Issuer element MUST have a value - TR pag. 9',
            **error_kwargs
        )

        entitydescriptor = self.md.xpath('//EntityDescriptor')
        entityid = entitydescriptor[0].get('entityID')
        self._assertEqual(
            e.text, entityid,
            "The Issuer's value MUST be equal to entityID - TR pag. 9",
            **error_kwargs
        )

        for attr in ['Format', 'NameQualifier']:
            self._assertTrue(
                (attr in e.attrib),
                f'The {attr} attribute MUST be present - TR pag. 9',
                **error_kwargs
            )

            value = e.get(attr)

            self._assertIsNotNone(
                value,
                f'The {attr} attribute MUST have a value - TR pag. 9',
                **error_kwargs
            )

            if attr == 'Format':
                exp = 'urn:oasis:names:tc:SAML:2.0:nameid-format:entity'
                self._assertEqual(
                    value,
                    exp,
                    f'The {attr} attribute MUST be {exp} - TR pag. 9',
                    **error_kwargs
                )
        return self.is_ok(f'{self.__class__.__name__}.test_Issuer')

    def test_NameIDPolicy(self):
        '''Test the compliance of NameIDPolicy element'''

        e = self.doc.xpath('//AuthnRequest/NameIDPolicy')

        desc = [etree.tostring(ent).decode() for ent in e if e]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertTrue(
            (len(e) == 1),
            'One NameIDPolicy element MUST be present - TR pag. 9',
            **error_kwargs
        )

        if not e:
            return self.is_ok(f'{self.__class__.__name__}.test_AuthnRequest')
        else:
            e = e[0]

        self._assertTrue(
            ('AllowCreate' not in e.attrib),
            'The AllowCreate attribute MUST not be present - AV n.5 ',
            **error_kwargs
        )

        attr = 'Format'
        self._assertTrue(
            (attr in e.attrib),
            f'The {attr} attribute MUST be present - TR pag. 9',
            **error_kwargs
        )

        value = e.get(attr)

        self._assertIsNotNone(
            value,
            f'The {attr} attribute MUST have a value - TR pag. 9',
            **error_kwargs
        )

        if attr == 'Format':
            exp = 'urn:oasis:names:tc:SAML:2.0:nameid-format:transient'
            self._assertEqual(
                value,
                exp,
                f'The {attr} attribute MUST be {exp} - TR pag. 9',
                **error_kwargs
            )
        return self.is_ok(f'{self.__class__.__name__}.test_NameIDPolicy')

    def test_Conditions(self):
        '''Test the compliance of Conditions element'''
        e = self.doc.xpath('//AuthnRequest/Conditions')

        desc = [etree.tostring(ent).decode() for ent in e if e]
        error_kwargs = dict(description = desc) if desc else {}

        if len(e) > 1:
            self._assertEqual(
                len(1),
                1,
                'Only one Conditions element is allowed - TR pag. 9',
                **error_kwargs
            )

        if len(e) == 1:
            e = e[0]
            for attr in ['NotBefore', 'NotOnOrAfter']:
                self._assertTrue(
                    (attr in e.attrib),
                    f'The {attr} attribute MUST be present - TR pag. 9',
                    **error_kwargs
                )

                value = e.get(attr)

                self._assertIsNotNone(
                    value,
                    f'The {attr} attribute MUST have a value - TR pag. 9',
                    **error_kwargs
                )

                self._assertTrue(
                    bool(constants.regex.UTC_STRING.search(value)),
                    f'The {attr} attribute MUST have avalid UTC string - TR pag. 9',
                    **error_kwargs
                )
        return self.is_ok(f'{self.__class__.__name__}.test_Conditions')

    def test_RequestedAuthnContext(self):
        '''Test the compliance of RequestedAuthnContext element'''

        e = self.doc.xpath('//AuthnRequest/RequestedAuthnContext')

        desc = [etree.tostring(ent).decode() for ent in e if e]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertEqual(
            len(e),
            1,
            'Only one RequestedAuthnContext element MUST be present - TR pag. 9',
            **error_kwargs
        )
        if e:
            e = e[0]

            attr = 'Comparison'
            self._assertTrue(
                (attr in e.attrib),
                f'The {attr} attribute MUST be present - TR pag. 10',
                **error_kwargs
            )

            value = e.get(attr)
            self._assertIsNotNone(
                value,
                f'The {attr} attribute MUST have a value - TR pag. 10',
                **error_kwargs
            )

            allowed = ['exact', 'minimum', 'better', 'maximum']
            self._assertIn(
                value,
                allowed,
                'Attribute not valid - TR pag. 10',
                description = f"The {attr} attribute MUST be one of [{', '.join(allowed)}]"
            )

            acr = e.xpath('./AuthnContextClassRef')
            self._assertEqual(
                len(acr),
                1,
                'Only one AuthnContexClassRef element MUST be present - TR pag. 9',
                description = [etree.tostring(_acr).decode() for _acr in acr]
            )

            if acr:
                acr = acr[0]
                self._assertIsNotNone(
                    acr.text,
                    'The AuthnContexClassRef element MUST have a value - TR pag. 9',
                    description = etree.tostring(acr)
                )

                self._assertTrue(
                    bool(constants.SPID_LEVEL_ALL.search(acr.text)),
                    'The AuthnContextClassRef element MUST have a valid SPID level - TR pag. 9 and AV n.5',
                    description = etree.tostring(acr)
                )
        return self.is_ok(f'{self.__class__.__name__}.test_RequestedAuthnContext')

    def test_Signature(self):
        '''Test the compliance of Signature element'''

        if not self.IS_HTTP_REDIRECT:
            sign = self.doc.xpath('//AuthnRequest/Signature')

            desc = [etree.tostring(ent).decode() for ent in sign if sign]
            error_kwargs = dict(description = desc) if desc else {}

            self._assertTrue((len(sign) == 1),
                             'The Signature element MUST be present - TR pag. 10',
                             **error_kwargs)

            method = sign[0].xpath('./SignedInfo/SignatureMethod')
            self._assertTrue((len(method) == 1),
                             'The SignatureMethod element MUST be present- TR pag. 10',
                             **error_kwargs)

            self._assertTrue(('Algorithm' in method[0].attrib),
                             'The Algorithm attribute MUST be present '
                             'in SignatureMethod element - TR pag. 10',
                             **error_kwargs)

            alg = method[0].get('Algorithm')
            self._assertIn(alg, constants.ALLOWED_XMLDSIG_ALGS,
                           'The signature algorithm MUST be valid - TR pag. 10',
                           description=
                                f"One of {', '.join(constants.ALLOWED_XMLDSIG_ALGS)}"
                          )  # noqa

            method = sign[0].xpath('./SignedInfo/Reference/DigestMethod')
            self._assertTrue((len(method) == 1),
                             'The DigestMethod element MUST be present',
                             **error_kwargs)

            self._assertTrue(('Algorithm' in method[0].attrib),
                             'The Algorithm attribute MUST be present '
                             'in DigestMethod element - TR pag. 10',
                             **error_kwargs)

            alg = method[0].get('Algorithm')
            self._assertIn(alg, constants.ALLOWED_DGST_ALGS,
                           (('The digest algorithm MUST be one of [%s] - TR pag. 10') %
                            (', '.join(constants.ALLOWED_DGST_ALGS))),
                           **error_kwargs)

            # save the grubbed certificate for future alanysis
            # cert = sign[0].xpath('./KeyInfo/X509Data/X509Certificate')[0]
            # dump_pem.dump_request_pem(cert, 'authn', 'signature', DATA_DIR)
        return self.is_ok(f'{self.__class__.__name__}.test_Signature')

    def test_RelayState(self):
        '''Test the compliance of RelayState parameter'''
        if ('RelayState' in self.params) and self.params.get('RelayState'):
            relaystate = self.params['RelayState']
            self._assertTrue(
                (relaystate.find('http') == -1),
                'RelayState MUST not be immediately intelligible - TR pag. 14 or pag. 15',
                description = relaystate
            )
        else:
            self._assertTrue(
                False, 'RelayState is missing - TR pag. 14 or pag. 15',
                description = 'Missing RelayState'
            )
        return self.is_ok(f'{self.__class__.__name__}.test_RelayState')

    def test_Scoping(self):
        '''Test the compliance of Scoping element'''

        e = self.doc.xpath('//AuthnRequest/Scoping')

        desc = [etree.tostring(ent).decode() for ent in e if e]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertEqual(
            len(e),
            0,
            'The Scoping element MUST not be present - AV n.5',
            **error_kwargs
        )
        return self.is_ok(f'{self.__class__.__name__}.test_Scoping')

    def test_RequesterID(self):
        '''Test the compliance of RequesterID element'''

        e = self.doc.xpath('//AuthnRequest/RequesterID')

        desc = [etree.tostring(ent).decode() for ent in e if e]
        error_kwargs = dict(description = desc) if desc else {}

        self._assertEqual(
            len(e),
            0,
            'The RequesterID  element MUST not be present - AV n.5',
            **error_kwargs
        )
        return self.is_ok(f'{self.__class__.__name__}.test_RequesterID')

    def test_profile_saml2core(self):
        self.test_xsd()
        self.test_AuthnRequest()
        self.test_Subject()
        self.test_Issuer()
        self.test_Conditions()
        self.test_Scoping()
        self.test_RequesterID()

    def test_profile_spid_sp(self):
        self.test_profile_saml2core()

        self.test_RelayState()
        self.test_Signature()
        self.test_xmldsig()
        self.test_AuthnRequest_SPID()
        self.test_AuthnRequest_SPID_extra()
        self.test_NameIDPolicy()
        self.test_RequestedAuthnContext()



import urllib

from lxml import etree

from . indicepa import get_indicepa_by_ipacode


class SpidSpMetadataCheckAG(object):

    def test_extensions_public_ag(self, ext_types=[
                    "//ContactPerson/Extensions/PublicServicesFullAggregator",
                    "//ContactPerson/Extensions/PublicServicesLightAggregator",
                    "//ContactPerson/Extensions/PrivateServicesFullAggregator",
                    "//ContactPerson/Extensions/PrivateServicesLightAggregator",
                    "//ContactPerson/Extensions/PublicServicesFullOperator",
                    "//ContactPerson/Extensions/PublicServicesLightOperator"],
                    must=False):

        for ext_type in ext_types:
            ctype = self.doc.xpath(ext_type)
            if must:
                self._assertTrue(
                    ctype,
                    f'The {ext_type} element MUST be present',
                )

            if ctype:
                self._assertFalse(
                    ctype[0].text,
                    f'The {ext_type.title()} element MUST be empty',
                )

        return self.is_ok(f'{self.__class__.__name__}.test_extensions_public_ag')

    def test_entityid_qs(self):
        """ The entityID MUST not contains the query-string part """

        entity_desc = self.doc.xpath('//EntityDescriptor')
        eid = entity_desc[0].get('entityID')

        qs = urllib.parse.splitquery(eid)

        self._assertFalse(
            qs,
            (f'The entityID MUST not contains the query-string part'),
             description = eid,
        )
        return self.is_ok(f'{self.__class__.__name__}.test_entityid_qs')

    def test_entityid_contains(self, value=''):
        """ The entityID MUST contain ... """

        entity_desc = self.doc.xpath('//EntityDescriptor')
        eid = entity_desc[0].get('entityID')
        self._assertTrue(
            value in eid,
            (f'The entityID MUST contain {value}'),
             description = eid,
        )
        return self.is_ok(f'{self.__class__.__name__}.test_entityid_contains')



import os
import logging
import shutil

from . import BASE_DIR

logger = logging.getLogger(__name__)


from jinja2 import (Environment,
                    FileSystemLoader,
                    select_autoescape)


def render_html_report(data:dict,
                       display_name:str,
                       template_search_path:str = f'{BASE_DIR}/html',
                       output_folder:str = './html_report'):
    loader = Environment(
                loader=FileSystemLoader(searchpath = template_search_path),
                autoescape=select_autoescape(['html'])
            )

    template = loader.get_template('index.html')

    page_fname = f'{output_folder}/index.html'
    try:
        os.mkdir(output_folder)
        shutil.copytree(f'{template_search_path}/static',
                        f'{output_folder}/static')
    except FileExistsError as e:
        logger.warning(e)

    f = open(page_fname, 'w')
    f.write(template.render(report_data = data,
                            display_name = display_name))
    f.close()



import re

from lxml import etree
from . constants import EMAIL_REGEXP


class SpidSpMetadataCheckPrivate(object):

    def test_Contacts_Priv(self):
        entity_desc = self.doc.xpath('//ContactPerson')

        ipacode = self.doc.xpath('//ContactPerson/Extensions/IPACode')
        self._assertFalse(
            ipacode,
            'The IPACode element MUST NOT be present',
            description = ipacode,
        )

        exts = self.doc.xpath(
            '//ContactPerson/Extensions/CessionarioCommittente'
        )
        self._assertTrue(
            (len(exts) == 1),
            ('The CessionarioCommittente element MUST be present'),
            description = exts,
        )

        if exts:
            ext = exts[0]
            tise = self.doc.xpath(
                '//ContactPerson/Extensions/CessionarioCommittente/TerzoIntermediarioSoggettoEmittente'
            )
            if tise:
                tise = tise[0]
                self._assertTrue(
                    tise.text,
                    'If the TerzoIntermediarioSoggettoEmittente element if present it MUST have a value',
                    description = tise,
                )

        return self.is_ok(f'{self.__class__.__name__}.test_Contacts_Priv')
